#!/usr/bin/env python

import os
import sys
import json
import numpy as np
import pandas as pd
from antinex_utils.log.setup_logging import build_colorized_logger
from antinex_utils.consts import SUCCESS
from antinex_utils.build_scaler_train_and_test_datasets import \
    build_scaler_train_and_test_datasets
from keras.models import Sequential
from keras.layers import Dense
from keras.wrappers.scikit_learn import KerasClassifier
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
import seaborn as sns


name = "antinex-scaler-django"
log = build_colorized_logger(name=name)


# noqa git clone https://github.com/jay-johnson/antinex-datasets.git /opt/antinex-datasets
dataset = os.getenv(
    "DATASET",
    ("/opt/antinex-datasets/v1/webapps/"
     "django/training-ready/v1_django_cleaned.csv"))
model_backup_file = os.getenv(
    "MODEL_BACKUP_FILE",
    "/tmp/{}-full-model.h5".format(
        name))
model_weights_file = os.getenv(
    "MODEL_WEIGHTS_FILE",
    "/tmp/{}-weights.h5".format(
        name))
model_json_file = os.getenv(
    "MODEL_JSON_FILE",
    "/tmp/{}-model.json".format(
        name))
model_image_file = os.getenv(
    "MODEL_IMAGE_FILE",
    "/tmp/{}-predictions-vs-correct.png".format(
        name))
footnote_text = os.getenv(
    "PLOT_FOOTNOTE",
    "AntiNex v1")
image_title = "{} - Predictions and Correct Predictions".format(
    name)
show_predictions = bool(os.getenv(
    "SHOW_PREDICTIONS",
    "1") == "1")


seed = 42
np.random.seed(seed)


def build_model(
        num_features,
        loss,
        optimizer,
        metrics):
    """build_model

    Build the Keras Deep Neural Network Model

    :param num_features: number of features
    :param loss: loss function to apply
    :param optimizer: optimizer to use
    :param metrics: list of metrics
    """
    log.info("building model")
    model = Sequential()
    model.add(
        Dense(
            250,
            input_dim=num_features,
            kernel_initializer="uniform",
            activation="relu"))
    model.add(
        Dense(
            1,
            kernel_initializer="uniform",
            activation="sigmoid"))

    log.info(("compiling loss={} optimizer={} metrics={}")
             .format(
                loss,
                optimizer,
                metrics))

    model.compile(
        loss=loss,
        optimizer=optimizer,
        metrics=metrics)

    return model
# end of build_model


features_to_process = [
    "idx",
    "arp_hwlen",
    "arp_hwtype",
    "arp_id",
    "arp_op",
    "arp_plen",
    "arp_ptype",
    "dns_default_aa",
    "dns_default_ad",
    "dns_default_an",
    "dns_default_ancount",
    "dns_default_ar",
    "dns_default_arcount",
    "dns_default_cd",
    "dns_default_id",
    "dns_default_length",
    "dns_default_ns",
    "dns_default_nscount",
    "dns_default_opcode",
    "dns_default_qd",
    "dns_default_qdcount",
    "dns_default_qr",
    "dns_default_ra",
    "dns_default_rcode",
    "dns_default_rd",
    "dns_default_tc",
    "dns_default_z",
    "dns_id",
    "eth_id",
    "eth_type",
    "icmp_addr_mask",
    "icmp_code",
    "icmp_gw",
    "icmp_id",
    "icmp_ptr",
    "icmp_seq",
    "icmp_ts_ori",
    "icmp_ts_rx",
    "icmp_ts_tx",
    "icmp_type",
    "icmp_unused",
    "ip_id",
    "ip_ihl",
    "ip_len",
    "ip_tos",
    "ip_version",
    "ipv6_fl",
    "ipv6_hlim",
    "ipv6_nh",
    "ipv6_plen",
    "ipv6_tc",
    "ipv6_version",
    "ipvsix_id",
    "pad_id",
    "tcp_dport",
    "tcp_fields_options.MSS",
    "tcp_fields_options.NOP",
    "tcp_fields_options.SAckOK",
    "tcp_fields_options.Timestamp",
    "tcp_fields_options.WScale",
    "tcp_id",
    "tcp_seq",
    "tcp_sport",
    "udp_dport",
    "udp_id",
    "udp_len",
    "udp_sport"
]

if not os.path.exists(dataset):
    log.error(("Failed to find dataset={}")
              .format(
                    dataset))

log.info(("loading dataset={}")
         .format(
            dataset))
predict_feature = "label_value"
predict_rows_df = pd.read_csv(dataset)
log.info("converting to json")
predict_rows = predict_rows_df.to_json()

found_columns = list(predict_rows_df.columns.values)
num_features = len(found_columns) - 1
min_scaler_range = -1
max_scaler_range = 1
max_records = 100000
verbose = 1
test_size = 0.2
batch_size = 32
epochs = 15
num_splits = 2
loss = "binary_crossentropy"
optimizer = "adam"
metrics = [
    "accuracy"
]
histories = [
    "val_loss",
    "val_acc",
    "loss",
    "acc"
]
label_rules = {
    "labels": [
        "not_attack",
        "not_attack",
        "attack"
    ],
    "label_values": [
        -1,
        0,
        1
    ]
}

should_set_labels = False
labels_dict = {}
if "labels" in label_rules and "label_values" in label_rules:
    label_rows = label_rules["label_values"]
    for idx, lidx in enumerate(label_rows):
        if len(label_rules["labels"]) >= idx:
            should_set_labels = True
            labels_dict[str(lidx)] = \
                label_rules["labels"][idx]
# end of compiling labels dictionary


# define a function with no arguments for Keras Scikit-Learn API to work
# https://keras.io/scikit-learn-api/
def set_model():
    """set_model"""
    return build_model(
        num_features=num_features,
        loss=loss,
        optimizer=optimizer,
        metrics=metrics)


scaler_res = build_scaler_train_and_test_datasets(
    label=name,
    train_features=features_to_process,
    test_feature=predict_feature,
    df=predict_rows_df,
    test_size=test_size,
    seed=seed,
    scaler_cast_to_type="float32",
    min_feature_range=min_scaler_range,
    max_feature_range=max_scaler_range)

if scaler_res["status"] != SUCCESS:
    log.error(("Failed to scaler train and test datasets for csv={}")
              .format(
                dataset))
    sys.exit(1)

scaler_res_data = scaler_res["data"]
datanode = {
    "X_train": scaler_res_data["x_train"],
    "Y_train": scaler_res_data["y_train"],
    "X_test": scaler_res_data["x_test"],
    "Y_test": scaler_res_data["y_test"]
}

predict_rows_df[predict_feature] = scaler_res_data["scaled_test_dataset"]
train_scaler_df = pd.DataFrame(
    scaler_res_data["scaled_train_dataset"],
    columns=features_to_process)
test_scaler_df = pd.DataFrame(
    scaler_res_data["scaled_test_dataset"],
    columns=[predict_feature])

sample_rows = train_scaler_df[features_to_process]
target_rows = test_scaler_df
num_samples = len(sample_rows.index)
num_target_rows = len(target_rows.index)

log.info(("scaled datasets ready "
          "x_train={} y_train={} "
          "x_test={} y_test={}")
         .format(
            len(datanode["X_train"]),
            len(datanode["Y_train"]),
            len(datanode["X_test"]),
            len(datanode["Y_test"])))

log.info("building KerasClassifier")

model = KerasClassifier(
    build_fn=set_model,
    epochs=epochs,
    batch_size=batch_size,
    verbose=verbose)

log.info(("fitting model epochs={} batch={} "
          "rows={} features={}")
         .format(
            epochs,
            batch_size,
            len(datanode["X_train"]),
            len(found_columns)))

history = model.fit(
    datanode["X_train"],
    datanode["Y_train"],
    validation_data=(
        datanode["X_test"],
        datanode["Y_test"]),
    epochs=epochs,
    batch_size=batch_size,
    shuffle=False,
    verbose=verbose)

log.info("building estimators")

estimators = []
estimators.append(
    ("standardize",
        StandardScaler()))
estimators.append(
    ("mlp",
        model))

log.info("building pipeline")
pipeline = Pipeline(estimators)

# noqa https://machinelearningmastery.com/multi-class-classification-tutorial-keras-deep-learning-library/
log.info(("starting classification "
          "StratifiedKFold "
          "splits={} seed={}")
         .format(
            num_splits,
            seed))

kfold = StratifiedKFold(
    n_splits=num_splits,
    random_state=seed)

log.info(("classification cross_val_score"))
results = cross_val_score(
    pipeline,
    datanode["X_train"],
    datanode["Y_train"],
    cv=kfold)
scores = [
    results.std(),
    results.mean()
]
accuracy = {
    "accuracy": results.mean() * 100
}

log.info(("classification accuracy={} samples={}")
         .format(
            accuracy["accuracy"],
            num_samples))

# make predictions
predictions = model.predict(
    sample_rows.values,
    verbose=verbose)

log.info(("classification confusion_matrix samples={} "
          "predictions={} target_rows={}")
         .format(
            num_samples,
            len(predictions),
            num_target_rows))

cm = confusion_matrix(
    target_rows.values,
    predictions)

log.info(("classification has confusion_matrix={} "
          "predictions={} target_rows={}")
         .format(
            cm,
            len(predictions),
            num_target_rows))
# end of confusion matrix

rounded = [round(x[0]) for x in predictions]
sample_predictions = []

log.info(("scores={} accuracy={} "
          "merging samples={} with predictions={} "
          "labels={}")
         .format(
            scores,
            accuracy.get("accuracy", None),
            len(sample_rows.index),
            len(rounded),
            labels_dict))

ridx = 0
for idx, row in predict_rows_df.iterrows():
    if len(sample_predictions) > max_records:
        log.info(("hit max={} predictions")
                 .format(
                    max_records))
        break
    new_row = json.loads(row.to_json())
    cur_value = rounded[ridx]
    if predict_feature in row:
        new_row["_original_{}".format(
                predict_feature)] = \
            int(row[predict_feature])
    else:
        new_row["_original_{}".format(
                predict_feature)] = \
            "missing-from-dataset"
    if cur_value != int(row[predict_feature]):
        new_row["prediction_status"] = 1
    else:
        new_row["prediction_status"] = 0

    new_row[predict_feature] = int(cur_value)

    if should_set_labels:
        new_row["label_name"] = \
            labels_dict[str(int(cur_value))]
    new_row["_row_idx"] = ridx
    new_row["_count"] = idx
    sample_predictions.append(new_row)
    ridx += 1
# end of merging samples with predictions

log.info(("creating merged_predictions_df from sample_predictions={}")
         .format(
            len(sample_predictions)))

merged_predictions_df = pd.DataFrame(
    sample_predictions)

if show_predictions:
    for row_num, row in merged_predictions_df.iterrows():
        log.info(("row={} original_{}={} predicted={}")
                 .format(
                    row_num,
                    predict_feature,
                    row["_original_{}".format(
                        predict_feature)],
                    row[predict_feature]))
    # end seeing original vs predicted
# if showing all predictions to log

log.debug("saving weights")

# there are some known issues saving weights:
# https://github.com/keras-team/keras/issues/4875
model.model.save_weights(model_weights_file)

log.debug("saving full keras model to file")

model.model.save(model_backup_file)

log.debug("saving keras model as json to file")

with open(
        model_json_file,
        "w") as json_file:
    json_file.write(model.model.to_json())

log.info(("neural network created "
          "merged_predictions_df.index={} columns={} "
          "with cross_val_score={} and accuracy={}")
         .format(
            len(merged_predictions_df.index),
            merged_predictions_df.columns.values,
            scores,
            accuracy.get("accuracy", None)))

log.info(("saved model_backup_file={} "
          "model_weights_file={} model_json_file={}")
         .format(
            model_backup_file,
            model_weights_file,
            model_json_file))

sns.set(font="serif")
sns.set_context(
    "paper",
    rc={
        "font.size": 12,
        "axes.titlesize": 12,
        "axes.labelsize": 10})
fig, ax = plt.subplots(
    figsize=(15.0, 10.0))

ax = sns.distplot(
    merged_predictions_df["prediction_status"])

ax.get_figure().text(
    0.90,
    0.01,
    footnote_text,
    va="bottom",
    fontsize=8,
    color="#888888")

# More seaborn examples:

# ax = sns.regplot(
#     x=merged_predictions_df[predict_feature],
#     y=merged_predictions_df["_original_{}".format(
#         predict_feature)],
#     marker="+")
# ax.get_figure().text(
#     0.90,
#     0.01,
#     footnote_text,
#     va="bottom",
#     fontsize=8,
#     color="#888888")

# g = sns.FacetGrid(
#     data=merged_predictions_df[[
#         predict_feature,
#         "_original_{}".format(
#             predict_feature)]],
#     size=15.0,
#     palette="Set2",
#     col=predict_feature,
#     hue=predict_feature)
# g.map(
#     sns.regplot,
#     predict_feature,
#     "_original_{}".format(
#         predict_feature))
# g.fig.suptitle(
#     image_title)
# g.fig.get_children()[-1].set_bbox_to_anchor(
#     (1.1, 0.5, 0, 0))
# g.fig.text(
#    0.90,
#    0.01,
#    footnote_text,
#    va="bottom",
#    fontsize=8,
#    color="#888888")

# g = sns.pairplot(
#    data=merged_predictions_df[[
#        predict_feature,
#        "_original_{}".format(
#            predict_feature)]],
#    diag_kind="kde")
# g.fig.suptitle(
#    image_title)
# g.fig.text(
#    0.90,
#    0.01,
#    footnote_text,
#    va="bottom",
#    fontsize=8,
#    color="#888888")

# sns.lmplot(
#    x=predict_feature,
#    y="_original_{}".format(
#        predict_feature),
#    hue=predict_feature,
#    data=merged_predictions_df,
#    markers=["o", "x"],
#    palette="Set1")

# g = sns.jointplot(
#     x=predict_feature,
#     y="_original_{}".format(
#         predict_feature),
#     data=merged_predictions_df,
#     kind="reg")

# g = sns.heatmap(
#     merged_predictions_df[[
#         predict_feature,
#         "_original_{}".format(
#             predict_feature)]].values.astype("int"),
#     annot=True,
#     annot_kws={
#        "size": 16})

# plt.show()
ax.get_figure().savefig(
    model_image_file)
# fig.savefig(
#    model_image_file)
# g.fig.savefig(
#    model_image_file)

log.info(("saved model_backup_file={} model_weights_file={} "
          "model_json_file={} image_file={}")
         .format(
            model_backup_file,
            model_weights_file,
            model_json_file,
            model_image_file))



#!/usr/bin/env python

import os
import sys
import datetime
import argparse
import json
import pandas as pd
from antinex_core.log.setup_logging import build_colorized_logger
from celery_connectors.publisher import Publisher

parser = argparse.ArgumentParser(description=("Launch a Predict Dataset "
                                              "Request into the AntiNex "
                                              "core"))
parser.add_argument(
    "-f",
    help=("request json file to use default: "
          "./training/scaler-regression.json"),
    required=False,
    dest="request_file")
parser.add_argument(
    "-d",
    help="debug",
    required=False,
    dest="debug",
    action="store_true")
args = parser.parse_args()

name = "predict-publisher"
log = build_colorized_logger(name=name)

log.info("{} - start".format(name))

request_file = "./training/scaler-regression.json"
if args.request_file:
    request_file = args.request_file

exchange_name = "webapp.predict.requests"
routing_key = "webapp.predict.requests"
queue_name = "webapp.predict.requests"
auth_url = "redis://localhost:6379/6"
serializer = "json"

if not os.path.exists(request_file):
    log.error(("Missing request file={}")
              .format(
                request_file))
    sys.exit(1)

req_data = None
with open(request_file, "r") as cur_file:
    req_data = json.loads(cur_file.read())

if not os.path.exists(request_file):
    log.error(("Did not find request data in file={}")
              .format(
                request_file))
    sys.exit(1)

# import ssl
# Connection("amqp://", login_method='EXTERNAL', ssl={
#            "ca_certs": '/etc/pki/tls/certs/something.crt',
#            "keyfile": '/etc/something/system.key',
#            "certfile": '/etc/something/system.cert',
#            "cert_reqs": ssl.CERT_REQUIRED,
#          })
#
ssl_options = {}
app = Publisher(
    name,
    auth_url,
    ssl_options)

if not app:
    log.error(("Failed to connect to broker={}")
              .format(
                  auth_url))
else:

    # Now send:
    now = datetime.datetime.now().isoformat()
    body = req_data
    body["created"] = now
    log.info("loading predict_rows")
    predict_rows_df = pd.read_csv(req_data["dataset"])
    predict_rows = []
    for idx, org_row in predict_rows_df.iterrows():
        new_row = json.loads(org_row.to_json())
        new_row["idx"] = len(predict_rows) + 1
        predict_rows.append(new_row)
    body["predict_rows"] = pd.DataFrame(predict_rows).to_json()

    log.info(("using predict_rows={}")
             .format(
                len(predict_rows)))

    log.info(("Sending msg={} "
              "ex={} rk={}")
             .format(
                str(body)[0:10],
                exchange_name,
                routing_key))

    # Publish the message:
    msg_sent = app.publish(
        body=body,
        exchange=exchange_name,
        routing_key=routing_key,
        queue=queue_name,
        serializer=serializer,
        retry=True)

    log.info(("End - {} sent={}")
             .format(
                name,
                msg_sent))
# end of valid publisher or not



#!/usr/bin/env python

import os
import sys
import json
import argparse
import pandas as pd
from antinex_utils.log.setup_logging import build_colorized_logger
from antinex_utils.utils import ppj


name = "convert-to-json"
log = build_colorized_logger(name=name)


parser = argparse.ArgumentParser(
            description=("Convert the last few "
                         "rows in a dataset to JSON"))
parser.add_argument(
    "-f",
    help=("dataset to use default /opt/antinex-datasets/v1/webapps/"
          "django/training-ready/v1_django_cleaned.csv"),
    required=False,
    dest="dataset")
parser.add_argument(
    "-b",
    help=("rows from bottom of dataset"),
    required=False,
    dest="rows_from_bottom")
parser.add_argument(
    "-d",
    help="enable debug",
    required=False,
    dest="debug",
    action="store_true")
args = parser.parse_args()

debug = False
dataset = ("/opt/antinex-datasets/v1/webapps/django/training-ready/"
           "v1_django_cleaned.csv")
bottom_row_idx = -5
if args.debug:
    debug = True
if args.dataset:
    dataset = args.dataset
if args.rows_from_bottom:
    bottom_row_idx = int(args.rows_from_bottom)
    if bottom_row_idx > 0:
        bottom_row_idx = -1 * bottom_row_idx

if not os.path.exists(dataset):
    log.error(("failed to find file={}")
              .format(
                dataset))
    sys.exit(1)

if debug:
    log.info(("reading dataset={}")
             .format(
                dataset))
df = pd.read_csv(
    dataset)
output_predict_rows = []

if debug:
    log.info(("building predict_rows={}")
             .format(
                -1 * bottom_row_idx))
for ridx, row in df.iloc[bottom_row_idx:].iterrows():
    new_row = json.loads(row.to_json())
    new_row["_dataset_index"] = ridx
    output_predict_rows.append(new_row)
# end of building rows from dataset

print(ppj(output_predict_rows))



import datetime
import json
import pandas as pd
from antinex_core.log.setup_logging import build_colorized_logger
from antinex_utils.utils import ppj
from antinex_utils.consts import SUCCESS
import antinex_utils.make_predictions


name = "antinex-prc"
log = build_colorized_logger(name=name)


class AntiNexProcessor:

    def __init__(
            self,
            name="antinex-prc",
            max_msgs=100,
            max_models=100):
        """__init__

        :param name: log label
        :param max_msgs: num msgs to save for replay debugging (FIFO)
        :param max_models: num pre-trained models to keep in memory (FIFO)
        """

        self.name = name
        log.info(("{} - INIT")
                 .format(
                     self.name))

        self.models = {}
        self.recv_msgs = []
        self.max_msgs = max_msgs
        self.max_models = max_models
    # end of __init__

    def handle_messages(
            self,
            body,
            message):
        """handle_messages

        :param body: body contents
        :param message: message object
        """

        should_ack = True
        should_reject = False
        should_requeue = False
        try:
            delivery_info = message.delivery_info
            log.info(("{} - msg with routing_key={}")
                     .format(
                        self.name,
                        delivery_info["routing_key"]))
            if delivery_info["routing_key"] == "webapp.predict.requests":
                try:
                    self.handle_predict_message(
                        req=body,
                        message=message)
                except Exception as f:
                    log.error(("{} - "
                               "failed handling PREDICT request with ex={}")
                              .format(
                                self.name,
                                f))
                # end of try/ex
                should_ack = True
            elif delivery_info["routing_key"] == "webapp.train.requests":
                try:
                    self.handle_train_message(
                        req=body,
                        message=message)
                except Exception as f:
                    log.error(("{} - "
                               "failed handling TRAIN request with ex={}")
                              .format(
                                self.name,
                                f))
                # end of try/ex
                should_ack = True
            else:
                log.error(("{} - misconfiguration error - consumed message "
                           "from exchange={} with routing_key={} acking")
                          .format(
                            self.name,
                            delivery_info["exchange"],
                            delivery_info["routing_key"]))
                should_ack = True
            # end of handling messages from multiple queues
        except Exception as e:
            log.error(("{} - failed to handle message={} body={} ex={}")
                      .format(
                        self.name,
                        message,
                        body,
                        e))
        # end of try/ex

        self.recv_msgs.append(body)
        self.cleanup_internals()

        if should_ack:
            message.ack()
        elif should_reject:
            message.reject()
        elif should_requeue:
            message.reject()
        else:
            log.error(("{} - acking message={} body={} by default")
                      .format(
                        self.name,
                        message,
                        body))
            message.ack()
        # end of handling for message pub/sub
    # end of handle_messages

    def handle_train_message(
            self,
            req,
            message):
        """handle_train_message

        :param req: body contents
        :param message: message object
        """
        log.info(("{} train msg={} "
                  "req={}")
                 .format(
                     self.name,
                     message.delivery_info,
                     str(req)[0:10]))

        model_name = str(
            req["label"]).strip().lstrip().lower()

        req["use_model_name"] = model_name

        if self.models.get(model_name, None) is not None:
            log.info(("re-training model={}")
                     .format(
                        model_name))

        self.run_train_and_predict(req)
    # end of handle_train_message

    def handle_predict_message(
            self,
            req,
            message):
        """handle_predict_message

        :param req: body contents
        :param message: message object
        """
        log.info(("{} predict msg={} "
                  "req={}")
                 .format(
                     self.name,
                     message.delivery_info,
                     str(req)[0:10]))

        model_name = str(
            req["label"]).strip().lstrip().lower()

        if self.models.get(model_name, None) is not None:
            log.info(("Running predictions with existing model={}")
                     .format(
                        model_name))
            req["use_model_name"] = model_name
            req["use_existing_model"] = self.models[model_name]["data"]
            self.run_train_and_predict(req)
        else:
            log.info(("{} model is not stored - training")
                     .format(
                        model_name))
            self.handle_train_message(
                req=req,
                message=message)
        # end of if can use model to predict or need to train
    # end of handle_predict_message

    def run_train_and_predict(
            self,
            req):
        """run_train_and_predict

        :param req: message dict consumed from a queue
        """

        log.info(("{} loading predict_rows into a df")
                 .format(
                    req["use_model_name"]))

        predict_df = pd.read_json(req["predict_rows"])
        show_model_json = False
        try:
            show_model_json = bool(int(req.get(
                "show_model_json",
                "0")) == 1)
        except Exception as e:
            show_model_json = False
            log.error(("{} - Set show_model_json to 0 or 1 ex={}")
                      .format(
                        req["label"],
                        e))
        ml_type = req["ml_type"].lower()
        predict_feature = req["predict_feature"]
        predictions = []
        res = antinex_utils.make_predictions.make_predictions(req)
        if res["status"] == SUCCESS:
            log.info(("{} - processing results")
                     .format(
                        req["label"]))
            res_data = res["data"]
            model = res_data["model"]
            acc_data = res_data["acc"]
            are_predictions_merged = res_data["are_predicts_merged"]
            predictions = res_data["sample_predictions"]

            accuracy = acc_data.get(
                "accuracy",
                None)
            if are_predictions_merged:
                log.info(("{} - processing merged predictions")
                         .format(
                            req["label"]))
                merge_df = res_data["merge_df"]
                model_predict_feature = "_predicted_{}".format(
                    predict_feature)
                if model_predict_feature not in merge_df:
                    log.error(("{} missing predicted feature={} "
                               "from res={}")
                              .format(
                                req["label"],
                                model_predict_feature,
                                res))
                    return res

                for idx, row in merge_df.iterrows():
                    log.info(("cur_sample={} - {}={} predicted={}")
                             .format(
                            idx,
                            predict_feature,
                            float(row[predict_feature]),
                            float(row[model_predict_feature])))
                # same as the merge method in antinex-utils
            else:
                for idx, node in enumerate(predictions):
                    label_name = None
                    if "label_name" in node:
                        label_name = node["label_name"]
                    org_feature = "_original_{}".format(
                                    predict_feature)
                    original_value = None
                    if org_feature in node:
                        original_value = node[org_feature]
                    if "regression" in ml_type:
                        log.info(("sample={} - {}={} predicted={}")
                                 .format(
                                    node["_row_idx"],
                                    predict_feature,
                                    float(original_value),
                                    float(node[predict_feature])))
                    elif "classification" in ml_type:
                        log.info(("sample={} - {}={} predicted={} label={}")
                                 .format(
                                    node["_row_idx"],
                                    predict_feature,
                                    original_value,
                                    node[predict_feature],
                                    label_name))
                    else:
                        log.info(("sample={} - {}={} predicted={}")
                                 .format(
                                    node["_row_idx"],
                                    predict_feature,
                                    original_value,
                                    node[predict_feature]))
            # end of predicting predictions

            if show_model_json:
                log.info(("{} made predictions={} model={} ")
                         .format(
                            req["label"],
                            len(predict_df.index),
                            ppj(json.loads(model.model.to_json()))))

            log.info(("{} made predictions={} found={} "
                      "accuracy={}")
                     .format(
                        req["label"],
                        len(predict_df.index),
                        len(res_data["sample_predictions"]),
                        accuracy))

            log.info(("{} - saving model={}")
                     .format(
                        req["label"],
                        req["use_model_name"]))

            self.models[req["use_model_name"]] = {
                "data": res_data,
                "created": datetime.datetime.now()
            }
        else:
            log.info(("{} failed predictions={}")
                     .format(
                        req["label"],
                        len(predict_df.index)))
        # end of if good train and predict

        return res
    # end of run_train_and_predict

    def cleanup_internals(
            self):
        """cleanup_internals"""

        if len(self.recv_msgs) > self.max_msgs:
            self.recv_msgs.pop(0, None)
        # end of cleanup message replay

        if len(self.models) > self.max_models:
            oldest_model_name = None
            oldest_model_date = None
            for midx, model_name in enumerate(self.models):
                model_node = self.models[model_name]
                if not oldest_model_date:
                    oldest_model_date = model_node["created"]
                    oldest_model_name = model_name
                else:
                    if model_node["created"] < oldest_model_date:
                        oldest_model_date = model_node["created"]
                        oldest_model_name = model_name
            # end of finding the oldest model to remove

            if oldest_model_name:
                log.info(("{} hit max_models={} deleting name={} date={}")
                         .format(
                            self.name,
                            self.max_models,
                            oldest_model_name,
                            oldest_model_date))
                del self.models[oldest_model_name]
                log.info(("{} num_models={}")
                         .format(
                            self.name,
                            len(self.models)))
            # end of deleting oldest model
        # end of clean up pre-trained models
    # end of cleanup_internals

    def show_diagnostics(
            self):
        """show_diagnostics"""
        log.info(("{} - models={}")
                 .format(
                    self.name,
                    self.models))
        for midx, m in enumerate(self.recv_msgs):
            log.info(("msg={} contents={}")
                     .format(
                        midx,
                        ppj(m)))
    # end of show_diagnostics

    def shutdown(
            self):
        """shutdown"""
        log.info(("{} - shutting down - start")
                 .format(
                    self.name))
        self.state = "SHUTDOWN"
        self.show_diagnostics()
        log.info(("{} - shutting down - done")
                 .format(
                    self.name))
    # end of shutdown

# end of AntiNexProcessor



from celery import Celery
from celery import signals
from antinex_core.log.setup_logging import build_colorized_logger
from antinex_utils.utils import ev
from celery_connectors.subscriber import Subscriber
from antinex_core.antinex_processor import AntiNexProcessor


# Disable celery log hijacking
# https://github.com/celery/celery/issues/2509
@signals.setup_logging.connect
def setup_celery_logging(**kwargs):
    pass


name = "antinex"
log = build_colorized_logger(name=name)


class AntiNexCore:

    def __init__(
            self,
            name="",
            broker_url=ev(
                "BROKER_URL",
                "redis://localhost:6379/6"),
            train_queue_name=ev(
                "TRAIN_QUEUE",
                "webapp.train.requests"),
            predict_queue_name=ev(
                "PREDICT_QUEUE",
                "webapp.predict.requests"),
            max_msgs=100,
            max_models=100):
        """__init__

        :param name: worker name
        :param broker_url: connection string to broker
        :param train_queue_name: queue name for training requests
        :param predict_queue_name: queue name for predict requests
        :param max_msgs: num msgs to save for replay debugging (FIFO)
        :param max_models: num pre-trained models to keep in memory (FIFO)
        """

        self.name = name
        log.info(("{} - INIT")
                 .format(
                     self.name))

        self.state = "INIT"
        self.broker_url = broker_url

        # Setup queues:
        self.train_queue_name = train_queue_name
        self.predict_queue_name = predict_queue_name

        self.queues = [
            self.train_queue_name,
            self.predict_queue_name
        ]

        # Subscribers
        self.all_queues_sub = None

        # SSL Celery options dict
        self.ssl_options = {}

        # http://docs.celeryproject.org/en/latest/userguide/calling.html#calling-retry  # noqa
        # allow publishes to retry for a time
        self.task_publish_retry_policy = {
            "interval_max": 1,
            "max_retries": 120,     # None - forever
            "interval_start": 0.1,
            "interval_step": 0.2}

        # Confirm publishes with Celery
        # https://github.com/celery/kombu/issues/572
        self.transport_options = {
            "confirm_publish": True}

        self.conn_attrs = {
            "task_default_queue": "antinex.worker.control",
            "task_default_exchange": "antinex.worker.control",
            # noqa http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-worker_prefetch_multiplier
            "worker_prefetch_multiplier": 1,  # consume 1 message at a time
            # noqa http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-worker_prefetch_multiplie
            "prefetch_count": 3,  # noqa consume 1 message at a time per worker (3 workers)
            # noqa http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-broker_heartbeat
            "broker_heartbeat": 240,  # in seconds
            # noqa http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-broker_connection_max_retries
            "broker_connection_max_retries": None,  # None is forever
            # noqa http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-task_acks_late
            "task_acks_late": True,  # noqa on consume do not send an immediate ack back
            "task_publish_retry_policy": self.task_publish_retry_policy}

        self.processor = AntiNexProcessor(
            name="{}-prc".format(
                    self.name),
            max_msgs=max_msgs,
            max_models=max_models)
    # end of __init__

    def start(
            self,
            app,
            ssl_options=None,
            task_retry_policy=None,
            transport_options=None,
            conn_attrs=None,
            queues=None,
            callback=None):
        """start

        :param app: Celery app
        :param ssl_options: ssl dictionary
        :param task_retry_policy: retry policy
        :param transport_options: transport config dict
        :param conn_attrs: connection dict
        :param queues: name of queues to consume messages
        :param callback: callback method when a message is consumed
        """

        log.info(("{} - start")
                 .format(
                    self.name))

        use_queues = queues
        if not use_queues:
            use_queues = self.queues

        use_ssl_options = ssl_options
        if not use_ssl_options:
            use_ssl_options = self.ssl_options

        use_retry_policy = task_retry_policy
        if not use_retry_policy:
            use_retry_policy = self.task_publish_retry_policy

        use_transport_options = transport_options
        if not use_transport_options:
            use_transport_options = self.transport_options

        use_conn_attrs = conn_attrs
        if not use_conn_attrs:
            use_conn_attrs = self.conn_attrs

        use_callback = callback
        if not use_callback:
            use_callback = self.processor.handle_messages

        log.info(("{} - start - creating subscriber")
                 .format(
                    self.name))

        self.all_queues_sub = Subscriber(
            self.name,
            self.broker_url,
            app,
            use_ssl_options,
            **use_conn_attrs)

        log.info(("{} - start - activating consumer for "
                  "queues={} callback={}")
                 .format(
                    self.name,
                    use_queues,
                    use_callback.__name__))

        self.all_queues_sub.consume(
            callback=use_callback,
            queues=use_queues,
            exchange=None,
            routing_key=None,
            prefetch_count=use_conn_attrs["prefetch_count"])

        self.state = "ACTIVE"

        log.info(("{} - start - state={} done")
                 .format(
                    self.name,
                    self.state))
    # end of start

    def show_diagnostics(
            self):
        """show_diagnostics"""

        self.processor.show_diagnostics()
    # end of show_diagnostics

    def shutdown(
            self):
        """shutdown"""
        log.info(("{} - shutting down - start")
                 .format(
                    self.name))
        self.state = "SHUTDOWN"
        self.show_diagnostics()
        self.processor.shutdown()
        log.info(("{} - shutting down - done")
                 .format(
                    self.name))
    # end of shutdown

# end of AntiNexCore


log.info("loading Celery app")
app = Celery()

broker_url = ev(
    "BROKER_URL",
    "redis://localhost:6379/6")
train_queue_name = ev(
    "TRAIN_QUEUE",
    "webapp.train.requests")
predict_queue_name = ev(
    "PREDICT_QUEUE",
    "webapp.predict.requests")
max_msgs = int(ev(
    "MAX_MSGS",
    "100"))
max_models = int(ev(
    "MAX_MODELS",
    "10"))

log.info("Creating antinex core")
core = AntiNexCore(
    name="core",
    broker_url=broker_url,
    train_queue_name=train_queue_name,
    predict_queue_name=predict_queue_name,
    max_msgs=max_msgs,
    max_models=max_models)
try:
    log.info("Starting antinex core")
    core.start(
        app=app)
except Exception as e:
    log.info(("Core hit exception={} shutting down")
             .format(
                e))
    core.shutdown()
    log.info(("canceling consumer to queue={}")
             .format(
                train_queue_name))
    app.control.cancel_consumer(train_queue_name)
    log.info(("canceling consumer to queue={}")
             .format(
                predict_queue_name))
    app.control.cancel_consumer(predict_queue_name)
# end of try/ex



import os
import sys
import warnings
import unittest

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


cur_path, cur_script = os.path.split(sys.argv[0])
os.chdir(os.path.abspath(cur_path))

install_requires = [
    "antinex-utils",
    "celery",
    "celery-connectors",
    "celery-loaders",
    "colorlog",
    "coverage",
    "docker-compose",
    "flake8>=3.4.1",
    "future",
    "h5py",
    "keras",
    "matplotlib",
    "numpy",
    "pandas",
    "pep8>=1.7.1",
    "pipenv",
    "pycodestyle",
    "pydocstyle",
    "pylint",
    "seaborn",
    "scikit-learn",
    "requests",
    "tables",
    "tensorflow",
    "tox",
    "tqdm",
    "unittest2",
    "mock"
]


if sys.version_info < (3, 5):
    warnings.warn(
        "Less than Python 3.5 is not supported.",
        DeprecationWarning)


# Do not import antinex_core module here, since deps may not be installed
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "antinex_core"))

setup(
    name="antinex-core",
    cmdclass={"test": PyTest},
    version="1.0.12",
    description=("AntiNex publisher-subscriber core for processing "
                 "training and prediction requests for deep neural "
                 "networks to detect network exploits using Keras "
                 "and Tensorflow in near real-time."),
    long_description=("AntiNex publisher-subscriber core for processing "
                      "training and prediction requests for deep neural "
                      "networks to detect network exploits using Keras "
                      "and Tensorflow in near real-time."),
    author="Jay Johnson",
    author_email="jay.p.h.johnson@gmail.com",
    url="https://github.com/jay-johnson/antinex-core",
    packages=[
        "antinex_core",
        "antinex_core.log"
    ],
    package_data={},
    install_requires=install_requires,
    test_suite="setup.antinex_core_test_suite",
    tests_require=[
        "pytest"
    ],
    scripts=[
        "./run-antinex-core.sh",
        "./publish_train_request.py",
        "./publish_predict_request.py",
        "./antinex_core/scripts/antinex-scaler-django.py",
        "./antinex_core/scripts/standalone-scaler-django.py",
        "./antinex_core/scripts/convert-bottom-rows-to-json.py",
        "./publish_regression_predict.py"
    ],
    use_2to3=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ])
